#
# Copyright 2018, 2020 Antoine Sanner
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
serial test
"""

import numpy as np
import scipy.optimize
from NuMPI.Optimization import LBFGS
from NuMPI.Optimization.Wolfe import (
    second_wolfe_condition,
    first_wolfe_condition
)
from NuMPI import MPI
import pytest
import test.Optimization.minimization_problems as mp

pytestmark = pytest.mark.skipif(
    MPI.COMM_WORLD.Get_size() > 1,
    reason="tests only serial funcionalities, please execute with pytest")


def my_print(*args):
    # print(*args) # uncomment this line to enable debug prints.
    pass


@pytest.mark.parametrize("Objective", [mp.Trigonometric])
@pytest.mark.parametrize("n", [10, 30])
def test_compare_scipy(Objective, n):
    x0 = Objective.startpoint(n)

    resLBGFS = LBFGS(Objective.f, x0, jac=Objective.grad, maxcor=5, gtol=1e-8,
                     maxiter=1000)
    assert resLBGFS.success
    # Patch
    # resScipy = scipy.optimize.minimize(lambda x: Objective.f(np.reshape(x,
    # (-1,1))),np.reshape(x0,(-1,)), jac = lambda x : np.reshape(
    # Objective.grad(np.reshape(x,(-1,1))),(-1,)),method='L-BFGS-B',tol = 1e-4)
    resScipy = scipy.optimize.minimize(Objective.f, np.reshape(x0, (-1,)),
                                       jac=Objective.grad, method="L-BFGS-B",
                                       options=dict(gtol=1e-8))
    assert resScipy.success

    np.testing.assert_allclose(np.reshape(resLBGFS.x, (-1,)), resScipy.x,
                               rtol=1e-2,
                               atol=1e-5)  # TODO: The location of the
    # minimum seems o be hard to find
    assert np.abs(Objective.f(resScipy.x) - Objective.f(resLBGFS.x)) < 1e-8


def test_3D():
    # Gaussian Potential

    # quadratic
    toPlot = False

    def ex_fun(x):
        "x should be an np array"
        # x.shape = (-1, 1)
        return np.sum(np.dot((x ** 2).flat, np.array([1, 4, 9])), axis=0)

    def ex_jac(x):
        return 2 * np.array((1 * x[0], 4 * x[1], 9 * x[2]))

    xg, yg = np.linspace(-5, 5, 51), np.linspace(-6, 6, 51)

    def mat_fun(x_g, x_):
        Z = np.zeros((xg.size, yg.size))

        for i in range(Z.shape[0]):
            for j in range(Z.shape[1]):
                Z[j, i] = ex_fun(np.array([xg[i], yg[j], 0]))
        return Z

    # plot
    if toPlot:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        X, Y = np.meshgrid(xg, yg)
        ax.contour(X, Y, mat_fun(xg, yg))
        fig.show()

        my_print(ex_fun(np.array((0, 0, 0))))
        fig2, ax2 = plt.subplots()
        for z in [0]:
            for y in [0, 1, 2]:
                ax2.plot(xg, [ex_fun(np.array((x, y, z))) for x in xg])
                ax2.plot(
                    xg, [ex_jac(np.array((x, y, z)))[0] for x in xg], '--')
        # plt.show(block = True)

    # Initial history:
    x_old = np.array([2., 1., -1.], dtype=float)
    x_old.shape = (-1, 1)
    # x_old.shape=(-1,1)
    grad_old = ex_jac(x_old)

    # line search
    reslinesearch = scipy.optimize.minimize_scalar(
        fun=lambda alpha: ex_fun(x_old - grad_old * alpha), bounds=(0, 2000),
        method="bounded")
    assert reslinesearch.success
    my_print("alpha {}".format(reslinesearch.x))
    x = x_old - grad_old * reslinesearch.x
    assert ex_fun(x) < ex_fun(x_old)
    # grad = ex_jac(x)
    my_print("x first_linesearch {}".format(x))
    # k = 1
    my_print("Wolfe")

    my_print("1st: {}".format(
        first_wolfe_condition(ex_fun, x_old, ex_jac, -grad_old,
                              reslinesearch.x, beta1=1e-4)))
    my_print("2nd {}".format(
        second_wolfe_condition(x_old, ex_jac, -grad_old, reslinesearch.x,
                               beta2=0.9)))

    resscipy = scipy.optimize.minimize(ex_fun, x, jac=ex_jac,
                                       options=dict(gtol=1e-10, ftol=0))
    my_print("scipy success: {}".format(resscipy.success))
    my_print("scipy nit {}".format(resscipy.nit))
    my_print("scipy result: {}".format(resscipy.x))

    res = LBFGS(ex_fun, x, jac=ex_jac, x_old=x_old, maxcor=3, maxiter=100,
                gtol=1e-10, ftol=0)
    assert res.success
    my_print("nit {}".format(res.nit))

    if toPlot:
        import matplotlib.pyplot as plt
        fig3d = plt.figure()
        ax3d = fig3d.add_subplot(111, projection='3d')

        for it, i in zip(res['iterates'], range(len(res['iterates']))):
            ax3d.plot(it.x[0], it.x[1], it.x[2], '+k')
            # ax3d.annotate(i, it.x)
        fig3d.show()
        plt.show(block=True)

    np.testing.assert_almost_equal(res.x, np.zeros(res.x.shape), decimal=5)


def test_quadratic_nD():
    # quadratic
    n = 50

    factors = np.random.random(n) + 0.1

    def ex_fun(x):
        "x should be an np array"
        # x.shape = (-1, 1)
        return np.sum(np.dot((x ** 2).flat, factors ** 2), axis=0)

    def ex_jac(x):
        return 2 * np.diag(factors).dot(x)

    # plot

    # Initial history:
    x_old = np.random.random(n)
    x_old.shape = (-1, 1)
    # x_old.shape=(-1,1)
    grad_old = ex_jac(x_old)

    # line search
    alpha, phi, phi0, derphi = scipy.optimize.linesearch.scalar_search_wolfe2(
        lambda alpha: ex_fun(x_old - grad_old * alpha),
        lambda alpha: np.dot(ex_jac(x_old - grad_old * alpha).T, -grad_old))
    assert derphi is not None
    x = x_old - grad_old * alpha
    assert ex_fun(x) < ex_fun(x_old)
    # grad = ex_jac(x)
    my_print("x first_linesearch {}".format(x))
    # k = 1

    resscipy = scipy.optimize.minimize(ex_fun, x, jac=ex_jac)
    my_print("schipy success: {}".format(resscipy.success))

    # my_print(x)
    # my_print(ex_jac(x))

    res = LBFGS(ex_fun, x, jac=ex_jac, x_old=x_old, maxcor=5, maxiter=100,
                g2tol=1e-10)
    my_print("nit {}".format(res.nit))

    if False:
        import matplotlib.pyplot as plt
        fig3d = plt.figure()
        ax3d = fig3d.add_subplot(111, projection='3d')

        for it, i in zip(res['iterates'], range(len(res['iterates']))):
            ax3d.plot(it.x[0], it.x[1], it.x[2], '+k')
            # ax3d.annotate(i, it.x)
        fig3d.show()
        plt.show(block=True)

    np.testing.assert_almost_equal(res.x, np.zeros((n, 1)), decimal=4)


@pytest.mark.parametrize("n", [3, 5, 10,
                               20])  # This has Problems with Linesearch at
# high number of points
def test_gaussian_nD(n):
    factors = np.random.random(n) + 0.5
    shift = (np.random.random(n) - 0.5) * 0.1
    shift.shape = (-1, 1)

    def ex_fun(x):
        "x should be an np array"
        # x.shape = (-1, 1)
        return - np.exp(
            - np.sum(np.dot(((x - shift) ** 2).flat, factors ** 2), axis=0))

    def ex_jac(x):
        return - 2 * np.diag(factors).dot(x - shift) * ex_fun(x)

    # plot

    # Initial history:
    x = np.random.random(n) - 0.5
    x.shape = (-1, 1)

    # resscipy = scipy.optimize.minimize(ex_fun, x, jac=ex_jac)
    # my_print("schipy success: {}".format(resscipy.success))

    res = LBFGS(ex_fun, x, jac=ex_jac, maxcor=2, gtol=1e-5, maxiter=10000)
    my_print("nit {}".format(res.nit))

    if False:
        import matplotlib.pyplot as plt
        fig3d = plt.figure()
        ax3d = fig3d.add_subplot(111, projection='3d')

        for it, i in zip(res['iterates'], range(len(res['iterates']))):
            ax3d.plot(it.x[0], it.x[1], it.x[2], '+k')
            # ax3d.annotate(i, it.x)
        fig3d.show()
        plt.show(block=True)

    np.testing.assert_allclose(res.x, shift, atol=1e-3)


def test_x2_xcosy():
    def ex_fun(x_):
        x = x_.reshape((-1, 1))
        return .5 * x[0, 0] ** 2 + x[0, 0] * np.cos(x[1, 0])

    def ex_jac(x_):
        x = x_.reshape((-1, 1))
        return np.array([[x[0, 0] + np.cos(x[1, 0])],
                         [-x[0, 0] * np.sin(x[1, 0])]])

    # plot

    # plt.show(block=True)
    ######
    # Initial history:
    x = np.array([[3], [-4]], dtype=float)

    # k = 1

    res = LBFGS(ex_fun, x, jac=ex_jac, gtol=1e-5, maxcor=5, maxiter=10000,
                linesearch_options=dict(c1=1e-4, c2=0.999))
    my_print("nit {}".format(res.nit))

    ref = scipy.optimize.minimize(ex_fun, x.reshape(-1),
                                  jac=lambda x: ex_jac(x).reshape(-1))

    assert ref.success

    if False:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        xg, yg = np.linspace(-5, 5, 51), np.linspace(-6, 6, 51)

        def mat_fun(xg, yg):
            Z = np.zeros((xg.size, yg.size))
            for i in range(Z.shape[0]):
                for j in range(Z.shape[1]):
                    Z[j, i] = ex_fun(np.array([xg[i], yg[j]]))
            return Z

        X, Y = np.meshgrid(xg, yg)
        plt.colorbar(ax.contour(X, Y, mat_fun(xg, yg)))
        fig.show()
        for it, i in zip(res['iterates'], range(len(res['iterates']))):
            ax.plot(it.x[0], it.x[1], '+k')
            ax.annotate(i, it.x)
        plt.show(block=True)

    np.testing.assert_allclose(res.x.reshape(-1), ref.x, atol=1e-16, rtol=1e-5)



#
# Copyright 2019-2020 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
This is a modification of a part of scipy.optimize.linesearch to accept
simulataneous evaluation of the function and the derivative

SCIPY LICENCE:

Copyright © 2001, 2002 Enthought, Inc.
All rights reserved.

Copyright © 2003-2013 SciPy Developers.
All rights reserved.
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
Neither the name of Enthought nor the names of the SciPy Developers may be
used to endorse or promote products derived from this software without
specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS”
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

"""

import numpy as np
from warnings import warn


class LineSearchWarning(RuntimeWarning):
    pass


def scalar_search_wolfe2(phi_derphi, phi0=None,
                         old_phi0=None, derphi0=None,
                         c1=1e-4, c2=0.9, amax=None, step=None,
                         extra_condition=None, maxiter=10):
    """Find alpha that satisfies strong Wolfe conditions.

    alpha > 0 is assumed to be a descent direction.

    Parameters
    ----------
    phi : callable f(x)
        Objective scalar function.
    derphi : callable f'(x), optional
        Objective function derivative (can be None)
    phi0 : float, optional
        Value of phi at s=0
    old_phi0 : float, optional
        Value of phi at previous point
    derphi0 : float, optional
        Value of derphi at s=0
    c1 : float, optional
        Parameter for Armijo condition rule.
    c2 : float, optional
        Parameter for curvature condition rule.
    amax : float, optional
        Maximum step size
    extra_condition : callable, optional
        A callable of the form ``extra_condition(alpha, phi_value)``
        returning a boolean. The line search accepts the value
        of ``alpha`` only if this callable returns ``True``.
        If the callable returns ``False`` for the step length,
        the algorithm will continue with new iterates.
        The callable is only called for iterates satisfying
        the strong Wolfe conditions.
    maxiter : int, optional
        Maximum number of iterations to perform

    Returns
    -------
    alpha_star : float or None
        Best alpha, or None if the line search algorithm did not converge.
    phi_star : float
        phi at alpha_star
    phi0 : float
        phi at 0
    derphi_star : float or None
        derphi at alpha_star, or None if the line search algorithm
        did not converge.

    Notes
    -----
    Uses the line search algorithm to enforce strong Wolfe
    conditions.  See Wright and Nocedal, 'Numerical Optimization',
    1999, pg. 59-60.

    For the zoom phase it uses an algorithm by [...].

    """

    if phi0 is None:
        phi0, derphi0 = phi_derphi(0.)

    alpha0 = 0
    if step is not None:
        alpha1 = step
    else:
        if old_phi0 is not None and derphi0 != 0:
            alpha1 = min(1.0, 1.01 * 2 * (phi0 - old_phi0) / derphi0)
        else:
            alpha1 = 1.0

    if alpha1 < 0:
        alpha1 = 1.0

    phi_a1, derphi_a1 = phi_derphi(alpha1)

    phi_a0 = phi0
    derphi_a0 = derphi0

    if extra_condition is None:
        def extra_condition(alpha, phi):
            return True

    for i in range(maxiter):
        if alpha1 == 0 or (amax is not None and alpha0 == amax):
            # alpha1 == 0: This shouldn't happen. Perhaps the increment has
            # slipped below machine precision?
            alpha_star = None
            phi_star = phi0
            phi0 = old_phi0
            derphi_star = None

            if alpha1 == 0:
                msg = 'Rounding errors prevent the line search from converging'
            else:
                msg = "The line search algorithm could not find a solution " \
                      + \
                      "less than or equal to amax: %s" % amax

            warn(msg, LineSearchWarning)
            break

        if (phi_a1 > phi0 + c1 * alpha1 * derphi0) or \
                ((phi_a1 >= phi_a0) and (i > 1)):
            alpha_star, phi_star, derphi_star = \
                _zoom(alpha0, alpha1, phi_a0,
                      phi_a1, derphi_a0, phi_derphi,
                      phi0, derphi0, c1, c2, extra_condition)
            break

        if (abs(derphi_a1) <= -c2 * derphi0):
            if extra_condition(alpha1, phi_a1):
                alpha_star = alpha1
                phi_star = phi_a1
                derphi_star = derphi_a1
                break

        if (derphi_a1 >= 0):
            alpha_star, phi_star, derphi_star = \
                _zoom(alpha1, alpha0, phi_a1,
                      phi_a0, derphi_a1, phi_derphi,
                      phi0, derphi0, c1, c2, extra_condition)
            break

        alpha2 = 2 * alpha1  # increase by factor of two on each iteration
        if amax is not None:
            alpha2 = min(alpha2, amax)
        alpha0 = alpha1
        alpha1 = alpha2
        phi_a0 = phi_a1
        derphi_a0 = derphi_a1
        phi_a1, derphi_a1 = phi_derphi(alpha1)

    else:
        # stopping test maxiter reached
        alpha_star = alpha1
        phi_star = phi_a1
        derphi_star = derphi_a1
        warn('The line search algorithm did not converge', LineSearchWarning)

    return alpha_star, phi_star, phi0, derphi_star


def _cubicmin(a, fa, fpa, b, fb, c, fc):
    """
    Finds the minimizer for a cubic polynomial that goes through the
    points (a,fa), (b,fb), and (c,fc) with derivative at a of fpa.

    If no minimizer can be found return None

    """
    # f(x) = A *(x-a)^3 + B*(x-a)^2 + C*(x-a) + D

    with np.errstate(divide='raise', over='raise', invalid='raise'):
        try:
            C = fpa
            db = b - a
            dc = c - a
            denom = (db * dc) ** 2 * (db - dc)
            d1 = np.empty((2, 2))
            d1[0, 0] = dc ** 2
            d1[0, 1] = -db ** 2
            d1[1, 0] = -dc ** 3
            d1[1, 1] = db ** 3
            [A, B] = np.dot(d1, np.asarray([fb - fa - C * db,
                                            fc - fa - C * dc]).flatten())
            A /= denom
            B /= denom
            radical = B * B - 3 * A * C
            xmin = a + (-B + np.sqrt(radical)) / (3 * A)
        except ArithmeticError:
            return None
    if not np.isfinite(xmin):
        return None
    return xmin


def _quadmin(a, fa, fpa, b, fb):
    """
    Finds the minimizer for a quadratic polynomial that goes through
    the points (a,fa), (b,fb) with derivative at a of fpa,

    """
    # f(x) = B*(x-a)^2 + C*(x-a) + D
    with np.errstate(divide='raise', over='raise', invalid='raise'):
        try:
            D = fa
            C = fpa
            db = b - a * 1.0
            B = (fb - D - C * db) / (db * db)
            xmin = a - C / (2.0 * B)
        except ArithmeticError:
            return None
    if not np.isfinite(xmin):
        return None
    return xmin


def _zoom(a_lo, a_hi, phi_lo, phi_hi, derphi_lo,
          phi_derphi, phi0, derphi0, c1, c2, extra_condition):
    """
    Part of the optimization algorithm in `scalar_search_wolfe2`.
    """

    maxiter = 10
    i = 0
    delta1 = 0.2  # cubic interpolant check
    delta2 = 0.1  # quadratic interpolant check
    phi_rec = phi0
    a_rec = 0
    while True:
        # interpolate to find a trial step length between a_lo and
        # a_hi Need to choose interpolation here.  Use cubic
        # interpolation and then if the result is within delta *
        # dalpha or outside of the interval bounded by a_lo or a_hi
        # then use quadratic interpolation, if the result is still too
        # close, then use bisection

        dalpha = a_hi - a_lo
        if dalpha < 0:
            a, b = a_hi, a_lo
        else:
            a, b = a_lo, a_hi

        # minimizer of cubic interpolant
        # (uses phi_lo, derphi_lo, phi_hi, and the most recent value of phi)
        #
        # if the result is too close to the end points (or out of the
        # interval) then use quadratic interpolation with phi_lo,
        # derphi_lo and phi_hi if the result is still too close to the
        # end points (or out of the interval) then use bisection

        if (i > 0):
            cchk = delta1 * dalpha
            a_j = _cubicmin(a_lo, phi_lo, derphi_lo, a_hi, phi_hi,
                            a_rec, phi_rec)
        if (i == 0) or (a_j is None) or (a_j > b - cchk) or (a_j < a + cchk):
            qchk = delta2 * dalpha
            a_j = _quadmin(a_lo, phi_lo, derphi_lo, a_hi, phi_hi)
            if (a_j is None) or (a_j > b - qchk) or (a_j < a + qchk):
                a_j = a_lo + 0.5 * dalpha

        # Check new value of a_j

        phi_aj, derphi_aj = phi_derphi(a_j)
        if (phi_aj > phi0 + c1 * a_j * derphi0) or (phi_aj >= phi_lo):
            phi_rec = phi_hi
            a_rec = a_hi
            a_hi = a_j
            phi_hi = phi_aj
        else:
            if abs(derphi_aj) <= -c2 * derphi0 and extra_condition(a_j,
                                                                   phi_aj):
                a_star = a_j
                val_star = phi_aj
                valprime_star = derphi_aj
                break
            if derphi_aj * (a_hi - a_lo) >= 0:
                phi_rec = phi_hi
                a_rec = a_hi
                a_hi = a_lo
                phi_hi = phi_lo
            else:
                phi_rec = phi_lo
                a_rec = a_lo
            a_lo = a_j
            phi_lo = phi_aj
            derphi_lo = derphi_aj
        i += 1
        if (i > maxiter):
            # Failed to find a conforming step size
            a_star = None
            val_star = None
            valprime_star = None
            break
    return a_star, val_star, valprime_star



#
# Copyright 2018, 2020 Antoine Sanner
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


"""
MPI-parallel writing of matrices in numpy's 'npy' format.
"""

from .. import MPI
import numpy as np
import struct
import abc

from numpy.compat import asstr
from numpy.lib.format import magic, MAGIC_PREFIX, _filter_header
from numpy.lib.utils import safe_eval


def save_npy(fn, data, subdomain_locations=None, nb_grid_pts=None,
             comm=MPI.COMM_WORLD):
    """

    Parameters
    ----------
    data : numpy array : data owned by the processor
    location : index of the first element of data within the global data
    nb_grid_pts : nb_grid_pts of the global data
    comm : MPI communicator

    Returns
    -------

    """
    data = np.asarray(data)
    ndims = len(data.shape)

    if ndims == 1:
        data = data.reshape((-1, 1))

    if subdomain_locations is None:
        subdomain_locations = (0, 0)
    elif ndims == 1:
        subdomain_locations = (subdomain_locations, 0)
    nb_subdomain_grid_pts = data.shape

    if nb_grid_pts is None:
        nb_grid_pts = nb_subdomain_grid_pts
    elif ndims == 1:
        nb_grid_pts = (nb_grid_pts, 1)

    fortran_order = np.isfortran(data)

    from numpy.lib.format import dtype_to_descr, magic
    magic_str = magic(1, 0)

    arr_dict_str = str({
        'descr': dtype_to_descr(data.dtype),
        'fortran_order': fortran_order,
        'shape': (
            nb_grid_pts[0],) if ndims == 1 else nb_grid_pts
    })

    while (len(arr_dict_str) + len(magic_str) + 2) % 16 != 15:
        arr_dict_str += ' '
    arr_dict_str += '\n'
    header_len = len(arr_dict_str) + len(magic_str) + 2

    file = MPI.File.Open(comm, fn, MPI.MODE_CREATE | MPI.MODE_WRONLY)
    if comm.Get_rank() == 0:
        file.Write(magic_str)
        file.Write(np.int16(len(arr_dict_str)))
        file.Write(arr_dict_str.encode('latin-1'))

    if fortran_order:
        # the returned array will be in fortran_order.
        # the data is loaded in C_contiguous array but in a transposed manner
        # data.transpose() is called which swaps the shapes back again and
        # toggles C-order to F-order
        ix = 1
        iy = 0
    else:
        ix = 0
        iy = 1

    mpitype = MPI._typedict[data.dtype.char]
    filetype = mpitype.Create_vector(nb_subdomain_grid_pts[ix],
                                     # number of blocks  : length of data in
                                     # the non-contiguous direction
                                     nb_subdomain_grid_pts[iy],
                                     # length of block : length of data in
                                     # contiguous direction
                                     nb_grid_pts[iy]
                                     # stepsize: the data is contiguous in y
                                     # direction,
                                     # two matrix elements with same x
                                     # position are separated by ny in memory
                                     )  # create a type
    # see MPI_TYPE_VECTOR

    filetype.Commit()  # verification if type is OK
    file.Set_view(
        header_len + (subdomain_locations[ix] * nb_grid_pts[iy]
                      + subdomain_locations[iy]) * mpitype.Get_size(),
        filetype=filetype)
    if fortran_order:
        data = data.transpose()
    file.Write_all(data.copy())  # TODO: is the copy needed ?
    filetype.Free()

    file.Close()


class MPIFileTypeError(Exception):
    pass


class MPIFileIncompatibleResolutionError(Exception):
    pass


def mpi_read_bytes(file, nbytes):
    # allocate the buffer
    buf = np.empty(nbytes, dtype=np.int8)
    file.Read_all(buf)
    return buf.tobytes()


def load_npy(fn, subdomain_locations=None, nb_subdomain_grid_pts=None,
             comm=MPI.COMM_WORLD):
    file = MPIFileViewNPY(fn, comm)
    # if file.nb_grid_pts != nb_domain_grid_pts:
    #    raise MPIFileIncompatibleResolutionError(
    #        "nb_domain_grid_pts is {} but file nb_grid_pts is {}".format(
    #        nb_domain_grid_pts, file.nb_grid_pts))
    data = file.read(subdomain_locations, nb_subdomain_grid_pts)
    file.close()
    return data


class MPIFileView(metaclass=abc.ABCMeta):
    def __init__(self, fn, comm):
        self.fn = fn
        self.comm = comm
        # if hasattr read, it is a stream and it should not close the file
        self.close_file_on_error = not hasattr(fn, 'read')

    @abc.abstractmethod
    def _read_header(self):
        pass

    @abc.abstractmethod
    def read(self):
        pass


def make_mpi_file_view(fn, comm,
                       format=None):  # TODO: DISCUSS: oder als __init__ von
    # der MPIFileView Klasse ?
    readers = {
        "npy": MPIFileViewNPY
    }

    if format is not None:
        try:
            reader = readers[format]
        except KeyError:
            raise (ValueError(
                "Given format is not recognised, you should give {}".format(
                    readers.keys())))
        return reader(fn, comm)

    for reader in readers.values():
        try:
            return reader(fn, comm)
        except MPIFileTypeError:
            pass
    raise MPIFileTypeError(
        "No MPI filereader was able to open_topography the file {}".format(fn))


class MPIFileViewNPY(MPIFileView):
    """

    you may have a look at numpy.lib.format if you want to understand how
    this code works
    """

    def __init__(self, fn, comm):
        super().__init__(fn, comm)
        self._read_header()

    def detect_format(self):  # TODO: maybe useless
        try:
            self._read_header()
            return True
        except Exception:
            return False

    def _read_header(self):
        self.file = None
        try:
            magic_str = magic(1, 0)
            self.file = MPI.File.Open(self.comm, self.fn, MPI.MODE_RDONLY)  #
            magic_str = mpi_read_bytes(self.file, len(magic_str))
            if magic_str[:-2] != MAGIC_PREFIX:
                raise MPIFileTypeError(
                    "MAGIC_PREFIX missing at the beginning of file {}".format(
                        self.fn))

            version = magic_str[-2:]

            if version == b'\x01\x00':
                hlength_type = '<H'
            elif version == b'\x02\x00':
                hlength_type = '<I'
            else:
                raise MPIFileTypeError("Invalid version %r" % version)

            hlength_str = mpi_read_bytes(self.file,
                                         struct.calcsize(hlength_type))
            header_length = struct.unpack(hlength_type, hlength_str)[0]
            header = mpi_read_bytes(self.file, header_length)

            header = _filter_header(asstr(header))
            d = safe_eval(
                header)  # TODO: Copy from _read_array_header  with all the
            # assertions
            self.dtype = np.dtype(d['descr'])
            self.fortran_order = d['fortran_order']
            self.nb_grid_pts = d['shape']
            self.data_start = self.file.Get_position()
        except Exception as err:
            # FIXME! This should be handled through a resource manager
            if self.file is not None and self.close_file_on_error:
                self.file.Close()
            raise err

    def read(self, subdomain_locations=None, nb_subdomain_grid_pts=None):

        if subdomain_locations is None:
            subdomain_locations = (0, 0)
        if nb_subdomain_grid_pts is None:
            nb_subdomain_grid_pts = self.nb_grid_pts

        # Now how to start reading ?

        mpitype = MPI._typedict[self.dtype.char]

        if self.fortran_order:
            # the returned array will be in fortran_order.
            # the data is loaded in C_contiguous array but in a transposed
            # manner
            # data.transpose() is called which swaps the shapes back again and
            # toggles C-order to F-order
            ix = 1
            iy = 0
        else:
            ix = 0
            iy = 1

        # create a type
        filetype = mpitype.Create_vector(
            nb_subdomain_grid_pts[ix],
            # number of blocks  : length of data in the non-contiguous
            # direction
            nb_subdomain_grid_pts[iy],
            # length of block : length of data in contiguous direction
            self.nb_grid_pts[iy]
            # stepsize: the data is contiguous in y direction,
            # two matrix elements with same x position are separated by ny
            # in memory
        )

        filetype.Commit()  # verification if type is OK
        self.file.Set_view(
            self.data_start + (
                    subdomain_locations[ix] * self.nb_grid_pts[iy] +
                    subdomain_locations[iy]) * mpitype.Get_size(),
            filetype=filetype)

        data = np.empty((nb_subdomain_grid_pts[ix], nb_subdomain_grid_pts[iy]),
                        dtype=self.dtype)

        self.file.Read_all(data)
        filetype.Free()

        if self.fortran_order:
            data = data.transpose()

        return data

    def close(self):
        self.file.Close()



#
# Copyright 2018, 2020 Antoine Sanner
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import numpy as np

"""
These are Testfunctions extracted from

Mori, J. J., Garbow, B. S. & Hillstrom, K. E. Testing Unconstrained
 Optimization Software. 25 (1981).

In future I may use Scipy's example functions, see
 https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.rosen
.html#scipy.optimize.rosen
"""

import abc


class ObjectiveFunction(object, metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def f_grad(x):
        pass

    @classmethod
    def f(cls, x):
        return cls.f_grad(x)[0]

    @classmethod
    def grad(cls, x):
        return cls.f_grad(x)[1]

    @classmethod
    def plot_2D(cls):
        import matplotlib.pyplot as plt
        from helpers.plot_helpers import draw_npArrow2D
        fun = cls.f

        xg, yg = np.linspace(*cls.bounds, 51), np.linspace(*cls.bounds, 51)

        def mat_fun(xg, yg, fun):
            Z = np.zeros((xg.size, yg.size))

            for i in range(Z.shape[0]):
                for j in range(Z.shape[1]):
                    Z[j, i] = fun(np.array([xg[i], yg[j]]))
            return Z

        X, Y = np.meshgrid(xg, yg)

        plt.colorbar(
            plt.pcolormesh(X, Y, np.log10(mat_fun(xg, yg, fun)))).set_label(
            r"log10($\phi$)")
        plt.plot(1 / 2, 1 / 2, '+')  # standart starting point
        ax = plt.gca()
        for x in [np.array([[0], [1]]),
                  np.array([[-2], [-2]]),
                  np.array([[2], [-3]]),
                  np.array([[-3], [1]])]:
            draw_npArrow2D(ax, x, delta=cls.grad(x) / 10)

            eps = 1e-4
            # unit Vector with random Direction
            phi = (np.random.random(1) * 2 * np.pi).item()
            u = np.array([[np.cos(phi)], [np.sin(phi)]])

            print((cls.f(x + u * eps) - cls.f(x)) / eps)
            print(cls.grad(x).T @ u)

        plt.show(block=True)


class Trigonometric(ObjectiveFunction):
    """
    all values of n are allowed

    :param x: 1d array
    :return:
    """
    bounds = (-np.pi, np.pi)

    @staticmethod
    def f_grad(x_):
        n = x_.size
        old_shape = x_.shape
        x = np.reshape(x_, (-1, 1))
        idxVector = np.reshape(np.arange(1, n + 1), (-1, 1))
        f = n - np.sum(np.cos(x)) + idxVector * (1 - np.cos(x)) - np.sin(x)

        jac = np.reshape(np.sin(x), (1, -1)) + np.diag(
            (-np.cos(x) + idxVector * np.sin(x)).flat)
        return np.sum(f ** 2).item(), np.reshape(2 * jac.T @ f, old_shape)

    @staticmethod
    def startpoint(n):
        """
        standard starting point
        :param x:
        :return:
        """

        return np.reshape(np.ones(n) / n, (-1, 1))

    @staticmethod
    def plot_2D():
        import matplotlib.pyplot as plt
        from helpers.plot_helpers import draw_npArrow2D
        fun = Trigonometric.f

        xg, yg = np.linspace(-np.pi, np.pi, 51), np.linspace(-np.pi, np.pi, 51)

        def mat_fun(xg, yg, fun):
            Z = np.zeros((xg.size, yg.size))

            for i in range(Z.shape[0]):
                for j in range(Z.shape[1]):
                    Z[j, i] = fun(np.array([xg[i], yg[j]]))
            return Z

        X, Y = np.meshgrid(xg, yg)

        plt.colorbar(plt.contour(X, Y, mat_fun(xg, yg, fun)))
        plt.plot(1 / 2, 1 / 2, '+')  # standart starting point
        ax = plt.gca()
        for x in [np.array([[0], [1]]),
                  np.array([[-2], [-2]]),
                  np.array([[2], [-3]]),
                  np.array([[-3], [1]])]:
            draw_npArrow2D(ax, x, delta=Trigonometric.grad(x) / 10)

            eps = 1e-4
            # unit Vector with random Direction
            phi = (np.random.random(1) * 2 * np.pi).item()
            u = np.array([[np.cos(phi)], [np.sin(phi)]])

            print((Trigonometric.f(x + u * eps) - Trigonometric.f(x)) / eps)
            print(Trigonometric.grad(x).T @ u)

            Trigonometric.grad(x)

        plt.show(block=True)


class Extended_Rosenbrock(ObjectiveFunction):
    """

    https://github.com/cjtonde/optimize_rosenbrock/blob/master/src
    /optimize_rosenbrock.py

    n should be even

    :param x: 1d array
    :return:
    """
    bounds = (-4, 4)

    @staticmethod
    def f_grad(x):

        sumf2 = (np.sum(
            100 * (x[1::2] - x[:-1:2] ** 2) ** 2 + (1 - x[:-1:2]) ** 2)).item()

        grad = np.zeros_like(x)
        grad[1::2] = 200 * (x[1::2] - x[:-1:2] ** 2)
        grad[:-1:2] = -400 * x[:-1:2] * (x[1::2] - x[:-1:2] ** 2) - 2 * (
                1 - x[:-1:2])  # # d / dx2l-1

        return sumf2, grad

    @staticmethod
    def startpoint(n):
        """
        standard starting point
        :param n:
        :return: array of shape (1,n)
        """
        x0 = np.zeros(n, dtype=float)
        x0.shape = (-1, 1)
        x0[:-1:2] = -1.2
        x0[1::2] = 1

        return x0

    @staticmethod
    def minVal(*args):
        return 0

    @staticmethod
    def xmin(n):
        """
        Location of minimum according to

        Mori, J. J., Garbow, B. S. & Hillstrom, K. E. Testing Unconstrained
        Optimization Software. 25 (1981).

        This function not necessarily have only one Minimum in higher
        dimensional Space: see e.g. 10.1162/evco.2006.14.1.119

        :param n: number of DOF
        :return: array of size n
        """

        return np.ones((n, 1), dtype=float)


if __name__ == "__main__":
    Extended_Rosenbrock.plot_2D()
    Trigonometric.plot_2D()



#
# Copyright 2018, 2020 Antoine Sanner
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

# skip undefined names
# flake8: noqa F821


import numpy as np
import scipy.optimize

from NuMPI import MPI
from NuMPI.Tools import Reduction

from NuMPI.Optimization.linesearch import scalar_search_wolfe2

def donothing(*args, **kwargs):
    pass


def steepest_descent_wolfe2(x0, f_gradf, pnp=None, maxiter=10, **kwargs):
    """
    For first Iteration there is no history. We make a steepest descent
    satisfying strong Wolfe Condition
    :return:
    """
    # x_old.shape=(-1,1)
    phi0, grad0 = f_gradf(x0)
    grad0 = grad0.reshape((-1, 1))

    derphi0 = pnp.dot(grad0.T, -grad0).item()
    gradf = np.zeros_like(grad0)

    # stores the last evaluated gradf,
    # take care, the linesearch has to return the last evaluated value
    # TODO: Dangerous
    def _phi_phiprime(alpha):
        phi, gradf[...] = f_gradf(x0 - grad0 * alpha)
        phiprime = pnp.dot(gradf.T, -grad0).item()
        return phi, phiprime

    # like in lbfgsb
    # if (iter.eq. 0.and..not.boxed) then
    #    stp = min(one / dnorm, stpmx)

    alpha, phi, phi0, derphi = scalar_search_wolfe2(_phi_phiprime,
                                                    maxiter=maxiter,
                                                    phi0=phi0, derphi0=derphi0,
                                                    step=min(1., 1. / np.sqrt(
                                                        -derphi0)), **kwargs)

    assert derphi is not None, "Line Search in first steepest descent failed"
    x = x0 - grad0 * alpha

    return x, gradf, x0, grad0, phi, phi0, derphi


def LBFGS(fun, x, args=(), jac=None, x_old=None, maxcor=10, gtol=1e-5,
          ftol=2.2e-9, maxiter=15000,
          maxls=20, linesearch_options=dict(c1=1e-3, c2=0.9),
          pnp=Reduction(MPI.COMM_WORLD),
          store_iterates=None, printdb=donothing, callback=None, **options):
    r"""

    convergence if |grad|_{\infty} <= gtol or <= ftol is satisfied

    Parameters
    ----------
    fun
    x
    args
    jac
    x_old: initial guess
    maxcor: max number of history gradients stored
    gtol:
    ftol:
    maxiter
    maxls, default 20, as in scipy.optimize.fmin_l_bfgs_b
    linesearch_options: further options for the linesearch
    the result of the linesearch has to satisfy the strong wolfe condition
    See Wright and Nocedal, 'Numerical Optimization',p.34
    c1 parameter for the sufficient decrease condition
    c2 parameter for the curvature condition
    
    default values are choosen here to match the implementation in
    the Fortran subroutine L-BFGS-B 3.0 by
    Ciyou Zhu, Richard Byrd, and Jorge Nocedal

    See lbfgsb.f line 2497, with gtol=c2 and ftol=c1
    
    pnp
    store_iterates: stores each iterate of x, only debugging
    printdb
    options

    Returns
    -------

    """

    if callback is None:
        callback = donothing

    # print("jac = {}, type = {}".format(jac, type(jac)))
    if jac is True:
        def fun_grad(x):
            """
            reshapes the gradient in a convenient form
            Parameters
            ----------
            x

            Returns
            -------

            """
            f, grad = fun(x)
            return f, grad.reshape((-1, 1))

    elif jac is False:
        raise NotImplementedError(
            "Numerical evaluation of gradient not implemented")
    else:
        # function and gradient provided sepatately

        def fun_grad(x):
            """
            evaluates function and grad consequently (important, see issue #13)
            and reshapes the gradient in a convenient form
            Parameters
            ----------
            x

            Returns
            -------

            """
            # order matte
            f = fun(x)
            grad = jac(x).reshape((-1, 1))
            return f, grad

    # user can provide x in the shape of his convenience
    # but we will work with the shape (-1, 1)
    original_shape = x.shape

    x = x.reshape((-1, 1))

    if x_old is None:
        x_old = x.copy()
        x, grad, x_old, grad_old, phi, phi_old, derphi = \
            steepest_descent_wolfe2(x_old, fun_grad, pnp=pnp, maxiter=maxls,
                                    **linesearch_options)
    else:
        phi_old, grad_old = fun_grad(
            x_old)  # phi_old is never used, except for the convergence
        # criterion
        phi, grad = fun_grad(x)

    # full history of x is sored here if wished
    iterates = list()
    k = 1

    # n = x.size  # number of degrees of freedom
    gamma = 1

    Slist = []  # history of the steps of x
    Ylist = []  # history of gradient differences
    R = np.zeros((0, 0))

    grad2 = pnp.sum(grad ** 2)

    alpha = 0  # line search step size

    # Start loop
    # printdb(k)
    while True:
        # Update Sk,Yk
        # print("k= {}".format(k))
        if k > maxcor:
            # S = np.roll(S, -1)
            # S[:, -1] = (x - x_old).flat

            Slist[:-1] = Slist[1:]
            Slist[-1] = (x - x_old)

            # Y = np.roll(Y, -1)
            # Y[:, -1] = (grad - grad_old).flat

            Ylist[:-1] = Ylist[1:]
            Ylist[-1] = (grad - grad_old)

        else:
            # S = np.hstack([S, x - x_old])
            Slist.append(x - x_old)
            # Y = np.hstack([Y, grad - grad_old])
            Ylist.append(grad - grad_old)

        # 2.
        grad2prev = grad2
        grad2 = pnp.sum(np.asarray(grad) ** 2)

        ######################
        # check if job is done
        # if ((grad2 < g2tol if g2tol is not None else True) and
        #        (pnp.max(np.abs(grad)) < gtol if gtol is not None else
        #        True) and
        #        ((phi - phi_old) / max((1,abs(phi),abs(phi_old))) <= ftol
        #        if ftol is not None else True)):
        callback(x)

        if (pnp.max(np.abs(grad)) < gtol):
            result = scipy.optimize.OptimizeResult({
                'success': True,
                'x': x.reshape(
                    original_shape),
                'fun': phi,
                'jac': grad.reshape(
                    original_shape),
                'nit': k,
                'message':
                    'CONVERGENCE: '
                    'NORM_OF_GRADIENT_<=_GTOL',
                'iterates': iterates
            })
            # if iterates:
            #    result['iterates'] = iterates
            return result

        if ((phi_old - phi) <= ftol * max((1, abs(phi), abs(phi_old)))):
            result = scipy.optimize.OptimizeResult({
                'success': True,
                'x': x.reshape(
                    original_shape),
                'fun': phi,
                'jac': grad.reshape(
                    original_shape),
                'nit': k,
                'message':
                    'CONVERGENCE: '
                    'REL_REDUCTION_OF_F_<=_FACTR*EPSMCH',
                'iterates': iterates
            })
            # if iterates:
            #    result['iterates'] = iterates
            return result

        if k > maxiter:
            result = scipy.optimize.OptimizeResult({
                'success': False,
                'x': x.reshape(
                    original_shape),
                'fun': phi,
                'jac': grad.reshape(
                    original_shape),
                'nit': k,
                'iterates': iterates
            })

            return result
        ###########
        # new iteration

        STgrad = np.array([pnp.dot(si.T, grad) for si in Slist]).reshape(-1, 1)
        YTgrad = np.array([pnp.dot(yi.T, grad) for yi in Ylist]).reshape(-1, 1)
        # STgrad = pnp.dot(S.T, grad)
        # YTgrad = pnp.dot(Y.T, grad)

        if k > maxcor:
            S_now_T_grad_prev = np.roll(STgrad_prev, -1)
            S_now_T_grad_prev[-1] = - alpha * gamma * grad2prev - alpha * (
                    STgrad_prev.T.dot(p1) + gamma * YTgrad_prev.T.dot(p2))
        else:  # straightforward Version
            S_now_T_grad_prev = np.array(
                [pnp.dot(si.T, grad_old) for si in Slist]).reshape(-1, 1)

        if k > maxcor:
            R = np.roll(R, (-1, -1),
                        axis=(0, 1))  # mxm Matrix hold by all Processors
            R[-1, :] = 0
            STym1 = STgrad - S_now_T_grad_prev
            R[:, -1] = STym1.flat  # O(m x n)

        elif k == 1:
            # Rm = np.triu(pnp.dot(S.T, Y))
            R = np.triu(np.array(
                [[pnp.dot(si.T, yi).item() for yi in Ylist] for si in Slist]))
            # print(R)
        else:
            # Rm = np.vstack([Rm, np.zeros(k - 1)])
            # Rm = np.hstack([Rm, pnp.dot(S.T, Y[:, -1]).reshape(k, 1)])
            R = np.vstack([R, np.zeros(k - 1)])
            R = np.hstack([R, np.array(
                [pnp.dot(si.T, Ylist[-1]) for si in Slist]).reshape(k, 1)])
        if k > maxcor:
            D = np.roll(D, (-1, -1), axis=(0, 1))
            # D[-1,-1] = np.dot(Y[:,-1],Y[:,-1])# yk-1Tyk-1 # TOOPTIMIZE
            D[-1, -1] = R[-1, -1]
        else:
            # D = np.diag(np.einsum("ik,ik -> k", S, Y))
            D = np.diag(R.diagonal())
        assert D[-1, -1] > 0, "k = {}: ".format(k)  # Assumption of Theorem 2.2

        if k > maxcor:
            YTY = np.roll(YTY, (-1, -1), axis=(0, 1))
            YTY[-1, :-1] = YTY[:-1, -1] = (YTgrad[:-1] - YTgrad_prev[1:]).flat
            YTY[-1, -1] = grad2prev - grad2 + 2 * YTgrad[-1]
        else:
            # YTYm = pnp.dot(Y.T, Y)
            YTY = np.array(
                [[pnp.dot(yi1.T, yi2).item() for yi2 in Ylist] for yi1 in
                 Ylist])
        # Step 5.
        gamma = D[-1, -1] / YTY[
            -1, -1]  # n.b. D[-1,-1] = sk-1T yk-1 = yk-1T sk-1

        # Step 6. and 7. together
        Rinv = np.linalg.inv(R)
        RiSg = Rinv.dot(STgrad)
        p1 = Rinv.T.dot(D + gamma * YTY).dot(RiSg) - gamma * Rinv.T.dot(YTgrad)
        p2 = - RiSg  # TODO

        # temphstack=np.hstack([S, gamma * Y])

        # Hgradm = gamma * grad + S.dot(p1)  + gamma * Y.dot(p2)
        Hgrad = gamma * grad
        for si, yi, p1i, p2i in zip(Slist, Ylist, p1.flat, p2.flat):
            Hgrad += si * p1i.item() + gamma * yi * p2i.item()

        phi_old = float(phi)
        # printdb("Linesearch: ")
        grad_old[:] = grad
        x_old[:] = x

        def _phi_phiprime(alpha):
            phi, grad[...] = fun_grad(x - Hgrad * alpha)
            phiprime = pnp.dot(grad.T, -Hgrad).item()
            return phi, phiprime

        # TODO: oldphi0: is it allowed to stay outside of the search
        #  direction ?
        alpha, phi, phi0, derphi = scalar_search_wolfe2(
            _phi_phiprime,
            phi0=phi,
            derphi0=pnp.dot(grad.T, -Hgrad).item(),
            maxiter=maxls,
            **linesearch_options)

        printdb("derphi: {}".format(derphi))
        assert derphi is not None, "line-search did not converge"

        x = x - Hgrad * alpha

        if store_iterates == 'iterate':
            iterate = scipy.optimize.OptimizeResult(
                {
                    'x': x.copy().reshape(original_shape),
                    'fun': phi,
                    'jac': grad.reshape(original_shape)
                })
            iterates.append(iterate)

        printdb("k = {}".format(k))
        k = k + 1

        STgrad_prev = STgrad.copy()
        YTgrad_prev = YTgrad.copy()



import numpy as np
import scipy.optimize as optim
from inspect import signature


def constrained_conjugate_gradients(fun, hessp, x0, gtol=1e-8,
                                    mean_value=None, residual_plot=False,
                                    maxiter=5000):
    """
    Implementation of constrained conjugate gradient algorithm as described in,
    I.A. Polonsky, L.M. Keer, Wear 231, 206 (1999).

    See also
    M.R. Hestenes, Conjugate Direction Methods in Optimization, Springer, New York, 1980, Chaps. 2, 3.
    B.N. Pshenichny, Yu.M. Danilin,
        Numerical Methods in Optimization Problems, Nauka, Moscow, 1975, Chap. 3, in Russian.
    G. Zoutendijk, Mathematical Programming Methods, North Holland, Amsterdam, 1976, Chap. 16.

    Application to contact problems with adhesion:
    Rey, V. et al. Comput. Mech. 60, 69–81 (2017)


    Parameters
    ----------
    fun : callable.
                The objective function to be minimized.
                            fun(x) -> float(energy),ndarray(gradient)
                where x is the input ndarray.
                Note that energy is never used, you can return a dummy value.

    hessp : callable
            Function to evaluate the hessian product of the objective.
            Hessp should accept either 1 argument (desscent direction) or
            2 arguments (x,descent direction).
                            hessp(des_dir)->ndarray
                                    or
                            hessp(x,des_dir)->ndarray
            where x is the input ndarray and des_dir is the descent direction.

    x0 : ndarray
         Initial guess.
         ValueError is raised if "None" is provided.

    gtol : float, optional
           Default value : 1e-8
           convergence criterion is max(abs) and norm2 of the projected
           gradient < gtol.

    mean_value :  float, optional
               If you want to apply the mean_value constraint then provide an
               float value to the mean_value.

    residual_plot : bool, optional
                    Generates a plot between the residual and iterations.

    maxiter : int, optional
              Default, maxiter=5000
              Maximum number of iterations after which the program will exit.

    Returns
    -------
    OptimizeResult  : scipy.optimize object.
        Attributes:
         success: bool
         x: x,
         jac: residual = gradient(x),
         nit: n_iterations,
         message: 'CONVERGENCE: NORM_OF_GRADIENT_<=_GTOL' or 'NO CONVERGENCE: MAXITERATIONS REACHED'

    References
    ----------
    ..[1] I.A. Polonsky, L.M. Keer, Wear 231, 206 (1999)
    """

    if not isinstance(mean_value, (type(None), int, float)):
        raise ValueError('Inappropriate type: {} for mean_value whereas a '
                         'float \
            or int is expected'.format(type(mean_value)))

    if not isinstance(residual_plot, bool):
        raise ValueError('Inappropriate type: {} for "residual_plot" whereas '
                         'a bool \
                         is expected'.format(type(residual_plot)))

    if x0 is not None:
        x = x0.copy()
        x = x.flatten()
        delta = 0
        G_old = 1
    else:
        raise ValueError('Input required for x0/initial value !!')

    gaps = np.array([])
    iterations = np.array([])

    des_dir = np.zeros(np.shape(x))

    if residual_plot:
        gaps = np.append(gaps, 0)
        iterations = np.append(iterations, 0)

    n_iterations = 1

    while (n_iterations < maxiter + 1):

        '''Mask to truncate the negative values'''
        mask_neg = x <= 0
        x[mask_neg] = 0.0

        '''Initial residual or GAP'''
        residual = fun(x)[1]

        mask_c = x > 0
        if mean_value is not None:
            residual = residual - np.mean(residual[mask_c])

        G = np.sum(residual[mask_c] ** 2)

        des_dir[mask_c] = -residual[mask_c] + delta * (G / G_old) * des_dir[
            mask_c]
        des_dir[np.logical_not(mask_c)] = 0
        G_old = G

        '''Calculating step-length alpha'''

        sig = signature(hessp)
        if len(sig.parameters) == 2:
            hessp_val = hessp(x, des_dir)
        elif len(sig.parameters) == 1:
            hessp_val = hessp(des_dir)
        else:
            raise ValueError('hessp function has to take max 1 arg (descent '
                             'dir) or 2 args (x, descent direction)')

        '''Here hessp_val is used as r_ij in original algorithm'''
        if mean_value is not None:
            hessp_val = hessp_val - np.mean(hessp_val[mask_c])

        if mask_c.sum() != 0:
            '''alpha is TAU from algorithm'''
            alpha = -np.sum(residual[mask_c] * des_dir[mask_c]) / np.sum(hessp_val[mask_c] * des_dir[mask_c])
        else:
            # TODO: does anything happen when alpha is 0 or is the algorithm just stuck ?
            alpha = 0.0

        if alpha < 0:
            print("it {} : hessian is negative along the descent direction. "
                  "You will probably need linesearch "
                  "or trust region".format(n_iterations))

        x[mask_c] += alpha * des_dir[mask_c]

        '''mask for contact'''
        mask_neg = x <= 0
        '''truncating negative values'''
        x[mask_neg] = 0.0

        mask_g = residual < 0
        mask_overlap = np.logical_and(mask_neg, mask_g)

        if mask_overlap.sum() == 0:
            delta = 1
        else:
            delta = 0
            x[mask_overlap] = x[mask_overlap] - alpha * residual[mask_overlap]

        if mean_value is not None:
            '''Take care of constraint a_x*a_y*sum(p_ij)=P0'''
            P = np.mean(x)
            x *= (mean_value / P)

        if residual_plot:
            iterations = np.append(iterations, n_iterations)
            if mask_c.sum() != 0:
                gaps = np.append(gaps, np.max(abs(residual[mask_c])))
            else:
                gaps = np.append(gaps, np.max(abs(residual)))

        n_iterations += 1
        res_convg = False
        assert np.logical_not(np.isnan(x).any())

        if n_iterations >= 3:
            '''If converged'''
            if mask_c.sum() != 0:
                if np.max(abs(residual[mask_c])) <= gtol:
                    res_convg = True
                else:
                    res_convg = False

            if res_convg:
                result = optim.OptimizeResult(
                    {
                        'success': True,
                        'x': x,
                        'jac': residual,
                        'nit': n_iterations,
                        'message': 'CONVERGENCE: NORM_OF_GRADIENT_<=_GTOL',
                        })
                if residual_plot:
                    import matplotlib.pyplot as plt
                    plt.plot(iterations, np.log10(gaps))
                    plt.xlabel('iterations')
                    plt.ylabel('residuals')
                    plt.show()

                return result

            elif (n_iterations >= maxiter - 1):
                '''If no convergence'''
                result = optim.OptimizeResult({
                    'success': False,
                    'x': x,
                    'jac': residual,
                    'nit': n_iterations,
                    'message': 'NO-CONVERGENCE: MAXITERATIONS REACHED',
                    })

                if residual_plot:
                    import matplotlib.pyplot as plt
                    plt.plot(iterations, np.log10(gaps))
                    plt.xlabel('iterations')
                    plt.ylabel('residuals')
                    plt.show()

                return result



#
# Copyright 2018, 2020 Antoine Sanner
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import numpy as np


def first_wolfe_condition(fun, x0, fprime, direction, alpha, beta1):
    """
    p. 268, 11.19

    Keyword Arguments:
    fun         -- objective function to minimize
    x0          -- initial guess for solution
    fprime      -- Jacobian (gradient)
    direction   -- search direction (column vec)
    alpha       -- step size
    beta1       -- lower wolfe bound
    """
    return (fun(x0 + alpha * direction) <= fun(x0) +
            alpha * beta1 * float(np.dot(fprime(x0).T, direction)))


def second_wolfe_condition(x0, fprime, direction, alpha, beta2):
    """
    p. 270, 11.21

    Keyword Arguments:
    x0        -- initial guess for solution
    fprime    -- Jacobian (gradient) of objective function
    direction -- search direction
    alpha     -- step size
    beta2     -- upper wolfe bound
    """
    return (float(np.dot(fprime(x0 + alpha * direction).T, direction)) >=
            beta2 * float(np.dot(fprime(x0).T, direction)))



#
# Copyright 2018, 2020 Antoine Sanner
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np

from .. import MPI


def get_dtypeInfo(dtype):
    if dtype.kind == 'i':
        return np.iinfo(dtype)
    if dtype.kind == 'f':
        return np.finfo(dtype)
    raise ValueError


class Reduction:
    def __init__(self, comm=None):
        if comm is None:
            self.comm = MPI.COMM_SELF
        else:
            self.comm = comm

    def sum(self, arr, *args, **kwargs):
        """
        take care that the input arrays have the same datatype on all
        Processors !

        when you specify an axis along which make the sum, be shure that it
        is the direction in which data is decomposed (for slab data
        decomposition) !

        pencil data decomposition is not implemented yet.

        Parameters
        ----------
        arr: numpy Array

        Returns
        -------
        scalar np.ndarray , the sum of all Elements of the Array over all
        the Processors
        """

        locresult = np.sum(arr, *args, **kwargs)
        result = np.zeros_like(locresult)
        # print("Proc{}: result.dtype={}, locresult.dtype={} arr.dtype={
        # }".format(self.comm.Get_rank(),result.dtype,locresult.dtype,
        # arr.dtype))
        mpitype = MPI._typedict[locresult.dtype.char]
        self.comm.Allreduce([locresult, mpitype], [result, mpitype],
                            op=MPI.SUM)
        if type(locresult) == type(result):  # TODO:Why is this done again ?
            return result
        else:
            return type(locresult)(result)

    def max(self, arr):
        """
        take care that the input arrays have the same datatype on all
        Processors !
        Parameters
        ----------
        arr: numpy float array, can be empty

        Returns
        -------
        np.array of size 1, the max value of arr over all arrays, if all are
        empty this is -np.inf

        """
        result = np.asarray(0, dtype=arr.dtype)

        absmin = get_dtypeInfo(arr.dtype).min  # most negative value that can
        #                                        be stored in this datatype
        mpitype = MPI._typedict[arr.dtype.char]
        self.comm.Allreduce([np.max(arr) if arr.size > 0 else np.array(
            [absmin], dtype=arr.dtype), mpitype],
                            [result, mpitype], op=MPI.MAX)
        # FIXME: use the max of the dtype because np.inf only float
        # TODO: Not elegant, but following options didn't work
        # self.comm.Allreduce(np.max(arr) if arr.size > 0 else np.array(
        # None, dtype=arr.dtype), result, op=MPI.MAX)
        # Here when the first array isn't empty it is fine, but otherwise
        # the result will be nan
        #
        # self.comm.Allreduce(np.max(arr) if arr.size > 0 else np.array([],
        # dtype=arr.dtype), result, op=MPI.MAX)
        # Her MPI claims that the input and output array have not the same
        # datatype
        return result

    def min(self, arr):
        """
        take care that the input arrays have the same datatype on all
        Processors !
        Parameters
        ----------
        arr: numpy float array, can be empty

        Returns
        -------
        np.array of size 1, the min value of arr over all arrays, if all are
        empty this is np.inf

        """
        result = np.asarray(0, dtype=arr.dtype)
        absmax = get_dtypeInfo(arr.dtype).max  # most positive value that can
        #                                        be stored in this datatype
        mpitype = MPI._typedict[arr.dtype.char]
        self.comm.Allreduce([np.min(arr) if arr.size > 0 else np.array(
            [absmax], dtype=arr.dtype), mpitype],
                            [result, mpitype], op=MPI.MIN)
        return result

    def dot(self, a, b):
        locresult = np.dot(a, b)
        result = np.zeros_like(locresult)
        mpitype = MPI._typedict[locresult.dtype.char]
        self.comm.Allreduce([locresult, mpitype], [result, mpitype],
                            op=MPI.SUM)
        return result

    def any(self, arr):
        result = np.array(False, dtype=bool)
        self.comm.Allreduce(np.array(np.any(arr)), result, op=MPI.LOR)
        return result.item()

    def all(self, arr):
        result = np.array(False, dtype=bool)
        self.comm.Allreduce(np.array(np.all(arr), dtype=bool), result,
                            op=MPI.LAND)
        return result.item()



#
# Copyright 2019 Lars Pastewka
#           2018-2019 Antoine Sanner
# 
# ### MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from setuptools import setup, find_packages

import versioneer

setup(
    name="NuMPI",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    package_data={'': ['ChangeLog.md']},
    include_package_data=True,
    packages=find_packages(),
    python_requires='>=3.5.0',
    install_requires=['numpy',
                      'scipy>=1.1.0',
                      ],
    # metadata for upload to PyPI
    author="Antoine Sanner",
    author_email="antoine.sanner@imtek.uni-freiburg.de",
    description="Numerical tools for MPI-parallelized code",
    license="MIT",
    test_suite='tests',
)



#
# Copyright 2018, 2020 Antoine Sanner
#           2019 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np
import scipy.optimize
import pytest
from NuMPI.Optimization import LBFGS
from NuMPI.Tools import Reduction

import test.Optimization.minimization_problems as mp
import test.Optimization.MPI_minimization_problems as mpi_mp


@pytest.fixture
def pnp(comm):
    return Reduction(comm)


@pytest.mark.parametrize("Objectiveclass", [mpi_mp.MPI_Quadratic])
def test_minimize_call_column(pnp, Objectiveclass):
    """
    Checks the compatibility with a scipy.optimize.minimze call. Note that one
    should not use this type of call when running in parallel

    Parameters
    ----------
    pnp
    Objectiveclass

    Returns
    -------

    """

    comm = pnp.comm
    n = 10 + 2 * comm.Get_size()
    Objective = Objectiveclass(n, pnp=pnp)
    result = scipy.optimize.minimize(Objective.f_grad, Objective.startpoint(),
                                     jac=True, method=LBFGS, options={
            "gtol": 1e-8, "ftol": 1e-20, "pnp": pnp
        })
    assert result.success
    assert pnp.max(abs(
        np.reshape(result.x, (-1,)) - np.reshape(Objective.xmin(), (-1,))
    )) / pnp.max(abs(Objective.startpoint() - Objective.xmin())) < 1e-5


@pytest.mark.parametrize("Objectiveclass", [mpi_mp.MPI_Quadratic])
def test_minimize_call_row(pnp, Objectiveclass):
    comm = pnp.comm
    n = 10 + 2 * comm.Get_size()
    Objective = Objectiveclass(n, pnp=pnp)
    result = scipy.optimize.minimize(
        Objective.f_grad,
        Objective.startpoint().reshape(-1),
        method=LBFGS, jac=True,
        options={"gtol": 1e-8, "ftol": 1e-20, "pnp": pnp})
    assert result.success, ""
    assert pnp.max(abs(
        np.reshape(result.x, (-1,)) - np.reshape(Objective.xmin(), (-1,))
    )) / pnp.max(abs(Objective.startpoint() - Objective.xmin())) < 1e-5


# TODO: test when jac is Bool

@pytest.mark.parametrize("shape", [(10,), (10, 1), (1, 10), (2, 4)])
def test_shape_unchanged(shape):
    """
    This test is only serial
    """
    size = np.prod(shape)
    Objective = mp.Extended_Rosenbrock

    x0 = Objective.startpoint(size).reshape(shape)

    result = LBFGS(Objective.f_grad, x0, jac=True, gtol=1e-10, ftol=1e-40,
                   pnp=np)

    assert result.success, ""
    np.testing.assert_allclose(np.reshape(result.x, (-1,)),
                               np.reshape(Objective.xmin(size), (-1,)),
                               rtol=1e-5)

    assert result.x.shape == shape, \
        "shape of result mismatch shape of startpoint"
    assert result.jac.shape == shape, \
        "shape of result jac mismatch shape of startpoint"


# FIXME: Implement
def test_multiple_tol():
    pass


def test_gtol():
    pass


def test_g2tol():
    pass


def test_ftol():
    pass


def test():
    pass
